import torch

from ykutil.multiprocess import run_in_parallel


def tensor_fill(tensor, fill_value, floatadd, extra_add=0):
    for t, f, fl in zip(tensor, fill_value, floatadd):
        t[:] = f + fl + extra_add
    return tensor


def test_fill_value():
    tensors = [torch.zeros(n) for n in range(4, 8)]
    extra_kwargs = {"extra_add": 1}
    list_kws = {
        "fill_value": [1, 2, 3, 4],
        "tensor": tensors,
        "floatadd": [0.1, 0.2, 0.3, 0.4],
    }
    res = run_in_parallel(tensor_fill, list_kws, 3, extra_kwargs)
    # for i in range(4):
    #     tensor_fill(tensors[i], list_kws["fill_value"][i], 1)
    print(tensors)


if __name__ == "__main__":
    test_fill_value()



import pandas as pd
from pathlib import Path


def expand_json_cols(df: pd.DataFrame, cols: list[str] | str) -> pd.DataFrame:
    if isinstance(cols, str):
        cols = [cols]
    for col in cols:
        df = pd.concat([df, pd.json_normalize(df[col])], axis=1)
        df = df.drop(columns=[col])
    return df


def df_to_csv(filename: str):
    path = Path(filename)
    if not path.exists():
        raise FileNotFoundError(f"File not found: {path}")

    suffixes = [s.lower() for s in path.suffixes]

    # Load dataframe based on file extension(s)
    if ".tsv" in suffixes:
        df = pd.read_csv(path, sep="\t")
    elif ".csv" in suffixes:
        df = pd.read_csv(path)
    elif ".parquet" in suffixes or ".pq" in suffixes:
        df = pd.read_parquet(path)
    elif ".feather" in suffixes:
        df = pd.read_feather(path)
    elif ".pkl" in suffixes or ".pickle" in suffixes:
        df = pd.read_pickle(path)
    elif ".jsonl" in suffixes or ".ndjson" in suffixes:
        df = pd.read_json(path, lines=True)
    elif ".json" in suffixes:
        df = pd.read_json(path)
    elif ".xlsx" in suffixes or ".xls" in suffixes:
        df = pd.read_excel(path)
    else:
        raise ValueError(f"Unsupported file extension for {path.name}")

    # Build output path by stripping all suffixes and appending .csv
    suffix_str = "".join(suffixes)
    base_name = path.name[: -len(suffix_str)] if suffix_str else path.name
    target_csv = path.with_name(base_name + ".csv")

    df.to_csv(target_csv, index=False)

    return str(target_csv)



import gc
import io
from typing import Optional

import torch
import torch.nn.functional as F


def rolling_window(a: torch.Tensor, size: int) -> torch.Tensor:
    """
    >>> from torch import Tensor
    >>> rolling_window(Tensor([1,2,3]), 2)
    tensor([[1., 2.],
            [2., 3.]])
    """
    size = a.shape[:-1] + (a.shape[-1] - size + 1, size)
    strides = a.stride() + (a.stride()[-1],)
    return torch.as_strided(a, size=size, stride=strides)


def find_all_subarray_poses(
    arr: torch.Tensor,
    subarr: torch.Tensor,
    end=False,
    roll_window: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """
    >>> from torch import Tensor
    >>> find_all_subarray_poses(Tensor([1,2,3,4,5,1,2,3]), Tensor([1,2]))
    tensor([0, 5])
    """
    if len(subarr) > len(arr):
        return torch.tensor([], dtype=torch.int64)
    if roll_window is None:
        roll_window = rolling_window(arr, len(subarr))
    poses = torch.nonzero(torch.all(roll_window == subarr, dim=1)).squeeze(1)
    if end and len(poses) > 0:
        poses += len(subarr)
    return poses


def disable_gradients(model: torch.nn.Module) -> torch.nn.Module:
    for param in model.parameters():
        param.requires_grad = False
    return model


def tensor_in(needles: torch.Tensor, haystack: torch.Tensor):
    """
    >>> from torch import Tensor
    >>> tensor_in(Tensor([7,3,4,1,9,12]), Tensor([9,4,13,3,0]))
    tensor([False,  True,  True, False,  True, False])
    """
    assert needles.ndim == haystack.ndim == 1
    if len(needles) == 0:
        return torch.tensor([], dtype=torch.bool)
    a1 = needles.expand((haystack.size(0), needles.size(0)))
    b1 = haystack.expand((needles.size(0), haystack.size(0))).t()
    return (a1 == b1).max(dim=0).values


def print_memory_info():
    t = torch.cuda.get_device_properties(0).total_memory
    r = torch.cuda.memory_reserved(0)
    a = torch.cuda.memory_allocated(0)
    fragment = r - a
    free = t - r
    print(
        f"\nMemory:\nTotal: {t}, Reserved: {r}, Allocated: {a}, Fragment: {fragment}, Free: {free}\n"
    )


def free_cuda_memory():
    gc.collect()
    torch.cuda.empty_cache()


def pad_along_dimension(tensors, dim, pad_value=0):
    """

    Pads a list of tensors along a specified dimension so they can be stacked.

    Args:
        tensors (list of torch.Tensor): List of tensors to pad.
        dim (int): The dimension along which to pad.

    Returns:
        torch.Tensor: A single stacked tensor with all inputs padded along the specified dimension.

    >>> pad_along_dimension([torch.randn(4,12), torch.randn(4,21), torch.randn(4,5)], dim=1).shape
    torch.Size([3, 4, 21])
    >>> pad_along_dimension([torch.randn(4,12,3), torch.randn(4,21,3), torch.randn(4,5,3)], dim=1).shape
    torch.Size([3, 4, 21, 3])
    >>> pad_along_dimension([torch.randn(12), torch.randn(21), torch.randn(5)], dim=0).shape
    torch.Size([3, 21])
    """
    # Determine the target size for the specified dimension
    max_size = max(t.size(dim) for t in tensors)

    # Create a function to apply padding to the tensor
    padded_tensors = []
    for t in tensors:
        pad_sizes = [0] * t.dim() * 2  # Create a list of zeros for all dimensions
        pad_sizes[len(pad_sizes) - 1 - dim * 2] = max_size - t.size(
            dim
        )  # Set the padding for the specified dimension
        padded_tensors.append(F.pad(t, pad_sizes, value=pad_value))

    # Stack the padded tensors
    return torch.stack(padded_tensors)


def serialize_tensor(t: torch.Tensor) -> bytes:
    buffer = io.BytesIO()
    torch.save(t, buffer)
    return buffer.getvalue()


def deserialize_tensor(b: bytes, map_location=None) -> torch.Tensor:
    buffer = io.BytesIO(b)
    return torch.load(buffer, map_location=map_location)


if __name__ == "__main__":
    # pad_along_dimension([torch.randn(4, 12), torch.randn(4, 21), torch.randn(4, 5)], 1)
    # pad_along_dimension([torch.randn(12), torch.randn(21), torch.randn(5)], 0)
    import doctest

    doctest.testmod()



from collections import defaultdict
from typing import Any, Type

import numpy as np
import torch


def describe_recursive(
    l: Any,
    types: defaultdict[int, set[str]],
    lengths: defaultdict[int, set[int]],
    arrays: defaultdict[int, set[tuple[Type, Type, tuple]]],
    dict_keys: defaultdict[int, set[Any]],
    depth=0,
):
    types[depth].add(type(l).__name__)
    if isinstance(l, list) or isinstance(l, tuple):
        lengths[depth].add(len(l))
        for el in l:
            describe_recursive(el, types, lengths, arrays, dict_keys, depth=depth + 1)
    elif isinstance(l, np.ndarray):
        arrays[depth].add((np.ndarray, l.dtype, l.shape))
    elif isinstance(l, torch.Tensor):
        arrays[depth].add((torch.Tensor, l.dtype, l.shape))
    elif isinstance(l, dict):
        dict_keys[depth].update(l.keys())


def describe_list(l: list | tuple, no_empty=True):
    assert isinstance(l, list) or isinstance(l, tuple), f"l is {type(l)}"
    types_by_depth = defaultdict(set)
    lenghts = defaultdict(set)
    arrays = defaultdict(set)
    dict_keys = defaultdict(set)
    describe_recursive(l, types_by_depth, lenghts, arrays, dict_keys, depth=0)

    obj = {
        "types": dict(types_by_depth),
        "lengths": dict(lenghts),
        "arrays": dict(arrays),
        "dict_keys": dict(dict_keys),
    }

    if no_empty:
        for k, v in tuple(obj.items()):
            if not v:
                del obj[k]

    return obj


def describe_array(arr: np.ndarray | list | torch.Tensor):
    if isinstance(arr, list):
        arr = np.array(arr)
    elif isinstance(arr, torch.Tensor):
        return f"shape: {arr.shape}, dtype: {arr.dtype}, min: {arr.min()}, mean: {arr.mean(dtype=float)}, max: {arr.max()}"
    assert isinstance(arr, np.ndarray)
    return f"shape: {arr.shape}, dtype: {arr.dtype}, min: {arr.min()}, mean: {arr.mean(dtype=float)}, max: {arr.max()}, std: {arr.std()}"



import os

from transformers import Trainer


def train_eval_and_get_metrics(trainer: Trainer, checkpoint=None):
    all_metrics = {
        "run_name": os.environ["WANDB_NAME"] if "WANDB_NAME" in os.environ else "tmp"
    }

    print("*** Train ***")
    train_result = trainer.train(resume_from_checkpoint=checkpoint)
    metrics = train_result.metrics
    trainer.log_metrics("train", metrics)
    trainer.save_metrics("train", metrics)
    trainer.save_state()
    all_metrics.update(metrics)

    print("*** Evaluate ***")
    metrics = trainer.evaluate(metric_key_prefix="eval")
    trainer.log_metrics("eval", metrics)
    trainer.save_metrics("eval", metrics)
    all_metrics.update(metrics)

    return all_metrics



import inspect
import logging
import os
from logging import FileHandler


class TruncatingFileHandler(FileHandler):
    def __init__(
        self,
        filename,
        encoding=None,
        delay=False,
        errors=None,
        max_size_mb=10,
        keep_rate=0.5,
    ):
        super().__init__(filename, "a", encoding, delay, errors)
        self.max_size_bytes = int(max_size_mb * 1024 * 1024)
        self.keep_rate = keep_rate

    def emit(self, record: logging.LogRecord) -> None:
        if self.should_truncate():
            self.truncate()
        return super().emit(record)

    def truncate(self):
        if os.path.exists(self.baseFilename):
            size = os.path.getsize(self.baseFilename)
            delete_first_n_bytes(self.baseFilename, int(size * (1 - self.keep_rate)))

    def should_truncate(self):
        if os.path.exists(self.baseFilename):
            size = os.path.getsize(self.baseFilename)
            return size > self.max_size_bytes
        return False


class RotatingFileHandle:
    """
    A file-like wrapper that automatically rotates log files when they exceed a size limit.
    When the file reaches max_size_mb, it keeps only the most recent portion of the content.

    This implementation uses a pipe to ensure all writes go through our rotation logic,
    preventing subprocess operations from bypassing the size checking.
    """

    def __init__(self, filepath: str, max_size_mb: float | int = 10):
        import threading
        import subprocess

        self.filepath = filepath
        self.max_size_bytes = int(max_size_mb * 1024 * 1024)

        # Create a pipe for communication
        self.read_fd, self.write_fd = os.pipe()

        # Create the actual file
        self.file_handle = open(filepath, "w")
        self.file_handle.close()  # Close it, we'll reopen as needed

        # Start a background thread to read from pipe and write to file
        self._stop_thread = False
        self._thread = threading.Thread(target=self._pipe_reader, daemon=True)
        self._thread.start()

    def _pipe_reader(self):
        """Background thread that reads from pipe and writes to file with rotation logic"""
        import select

        # Use binary mode for the pipe reader to handle all data correctly
        with os.fdopen(self.read_fd, "rb") as pipe_reader:
            while not self._stop_thread:
                # Use select to check if data is available (non-blocking)
                ready, _, _ = select.select([pipe_reader], [], [], 0.1)
                if ready:
                    try:
                        # Read available data in binary mode
                        data_bytes = pipe_reader.read(8192)  # Read in chunks
                        if data_bytes:
                            # Decode to string for writing to text file
                            data = data_bytes.decode("utf-8", errors="ignore")
                            self._write_to_file(data)
                        else:
                            # EOF reached
                            break
                    except Exception as e:
                        # Handle any errors in reading/writing
                        print(f"Error in pipe reader: {e}")
                        break

    def _write_to_file(self, data: str):
        """Write data to file with rotation logic"""
        if not data:
            return

        # Process data in smaller chunks to ensure we don't exceed size limits
        chunk_size = min(1024, self.max_size_bytes // 4)  # Use smaller chunks

        for i in range(0, len(data), chunk_size):
            chunk = data[i : i + chunk_size]
            if not chunk:
                continue

            # Check if adding this chunk would exceed the limit
            current_size = (
                os.path.getsize(self.filepath) if os.path.exists(self.filepath) else 0
            )
            chunk_size_bytes = len(chunk.encode("utf-8"))

            if current_size + chunk_size_bytes > self.max_size_bytes:
                self._rotate_file()

            # Write chunk to file
            with open(self.filepath, "a") as f:
                f.write(chunk)
                f.flush()

    def write(self, data: str):
        """Write data through the pipe"""
        if not self._stop_thread:
            os.write(self.write_fd, data.encode("utf-8"))

    def _rotate_file(self):
        """Keep the last 50% of the file when rotating"""
        try:
            if not os.path.exists(self.filepath):
                return

            # Read the file content
            with open(self.filepath, "r") as f:
                content = f.read()

            # Keep only the last 50% of content (by character count)
            keep_size = len(content) // 2
            new_content = content[-keep_size:] if keep_size > 0 else ""

            # Write back the truncated content
            with open(self.filepath, "w") as f:
                if new_content:
                    f.write("=== LOG ROTATED - OLDER ENTRIES REMOVED ===\n")
                    f.write(new_content)
        except Exception as e:
            # If rotation fails, just truncate the file
            try:
                with open(self.filepath, "w") as f:
                    f.write(f"=== LOG ROTATION ERROR: {e} - FILE TRUNCATED ===\n")
            except:
                pass

    def flush(self):
        """Flush the pipe"""
        try:
            os.fsync(self.write_fd)
        except:
            pass

    def close(self):
        """Close the handle and stop background thread"""
        self._stop_thread = True
        try:
            os.close(self.write_fd)
        except:
            pass
        if self._thread.is_alive():
            self._thread.join(timeout=1.0)

    def fileno(self):
        """Return the write end of the pipe for subprocess to use"""
        return self.write_fd

    def readable(self):
        return False

    def writable(self):
        return True

    def seekable(self):
        return False


def delete_first_n_bytes(filepath: str, n: int):
    """
    Delete the first n bytes from a file.

    Args:
        filepath (str): Path to the file to modify
        n (int): Number of bytes to delete from the beginning

    Returns:
        bool: True if successful, False if file doesn't exist or n <= 0
    """
    if n <= 0:
        return False

    if not os.path.exists(filepath):
        return False

    try:
        # Get file size first
        file_size = os.path.getsize(filepath)

        # If n is greater than or equal to file size, truncate the entire file
        if n >= file_size:
            with open(filepath, "w") as f:
                f.write("")
            return True

        # Read the file content after the first n bytes
        with open(filepath, "rb") as f:
            f.seek(n)  # Skip the first n bytes
            remaining_content = f.read()

        # Write the remaining content back to the file
        with open(filepath, "wb") as f:
            f.write(remaining_content)

        return True

    except Exception as e:
        log(f"Error deleting first {n} bytes from {filepath}: {e}", level="ERROR")
        return False


if not "log_util" in logging.Logger.manager.loggerDict:
    logger = logging.getLogger("log_util")
    logger.setLevel(logging.DEBUG)
    hl = logging.StreamHandler()
    hl.setLevel(logging.INFO)
    hl.setFormatter(logging.Formatter("%(asctime)s %(message)s", datefmt="%H:%M:%S"))
    logger.addHandler(hl)

level_map = {
    "WARN": logging.WARNING,
    "INFO": logging.INFO,
    "ERROR": logging.ERROR,
    "DEBUG": logging.DEBUG,
    "WARNING": logging.WARNING,
}


def add_file_handler(file_path, level=logging.INFO, max_bytes=None):
    file_dir = os.path.dirname(file_path)
    if not os.path.exists(file_dir):
        os.makedirs(file_dir, exist_ok=True)
    if isinstance(level, str):
        level = level_map[level.upper()]
    if max_bytes:
        fl = TruncatingFileHandler(file_path, max_size_mb=max_bytes)
    else:
        fl = logging.FileHandler(file_path, mode="a")
    fl.setLevel(level)
    fl.setFormatter(logging.Formatter("%(asctime)s %(message)s", datefmt="%H:%M:%S"))
    logger.addHandler(fl)


def log(*messages, level=logging.INFO):
    from termcolor import colored

    if type(level) == str:
        level = level_map[level.upper()]
    caller = inspect.stack()[1]
    msg = f'{os.path.basename(caller.filename)}:{caller.lineno} {" ".join(map(str, messages))}'
    if level == logging.ERROR:
        msg = colored(msg, "red")
    elif level == logging.WARNING:
        msg = colored(msg, "yellow")
    logger.log(level, msg)



import requests
import time


def wait_for_url(url: str, max_tries: int = 60, timeout: int = 2) -> bool:
    for _ in range(max_tries):
        try:
            resp = requests.get(url, timeout=timeout)
            if resp.status_code == 200:
                return True
        except Exception:
            pass
        time.sleep(3)
    return False



import yaml


class LiteralDumper(yaml.SafeDumper):
    def represent_str(self, data):
        """Force multiline strings to use the '|' block style."""
        if "\n" in data:  # Check if the string contains newlines
            return self.represent_scalar("tag:yaml.org,2002:str", data, style="|")
        return self.represent_scalar("tag:yaml.org,2002:str", data)


# Add the custom string representation method to the dumper
LiteralDumper.add_representer(str, LiteralDumper.represent_str)


def yaml_pretty_dump(data):
    return yaml.dump(
        data, Dumper=LiteralDumper, default_flow_style=False, allow_unicode=True
    )



import os
from typing import Dict

import torch
import transformers


def print_trainable_parameters(model_args, model):
    # Source https://github.com/artidoro/qlora/tree/main
    """
    Prints the number of trainable parameters in the model.
    """
    trainable_params = 0
    all_param = 0
    for _, param in model.named_parameters():
        all_param += param.numel()
        if param.requires_grad:
            trainable_params += param.numel()
    if model_args.bits == 4:
        trainable_params /= 2
    print(
        f"trainable params: {trainable_params} || "
        f"all params: {all_param} || "
        f"trainable: {100 * trainable_params / all_param}"
    )


def print_parameters_by_dtype(model):
    dtypes = {}
    for _, p in model.named_parameters():
        dtype = p.dtype
        if dtype not in dtypes:
            dtypes[dtype] = 0
        dtypes[dtype] += p.numel()
    total = 0
    for k, v in dtypes.items():
        total += v
    for k, v in dtypes.items():
        print(k, v, v / total)


def smart_tokenizer_and_embedding_resize(
    special_tokens_dict: Dict,
    tokenizer: transformers.PreTrainedTokenizer,
    model: transformers.PreTrainedModel,
):
    # Source https://github.com/artidoro/qlora/tree/main
    """Resize tokenizer and embedding.

    Note: This is the unoptimized version that may make your embedding size not be divisible by 64.
    """
    num_new_tokens = tokenizer.add_special_tokens(special_tokens_dict)
    model.resize_token_embeddings(len(tokenizer))

    if num_new_tokens > 0:
        input_embeddings_data = model.get_input_embeddings().weight.data
        output_embeddings_data = model.get_output_embeddings().weight.data

        input_embeddings_avg = input_embeddings_data[:-num_new_tokens].mean(
            dim=0, keepdim=True
        )
        output_embeddings_avg = output_embeddings_data[:-num_new_tokens].mean(
            dim=0, keepdim=True
        )

        input_embeddings_data[-num_new_tokens:] = input_embeddings_avg
        output_embeddings_data[-num_new_tokens:] = output_embeddings_avg


def find_all_linear_names(model, bits=32) -> list[str]:
    import bitsandbytes as bnb

    # Source https://github.com/artidoro/qlora/tree/main
    cls = (
        bnb.nn.Linear4bit
        if bits == 4
        else (bnb.nn.Linear8bitLt if bits == 8 else torch.nn.Linear)
    )
    lora_module_names = set()
    for name, module in model.named_modules():
        if isinstance(module, cls):
            names = name.split(".")
            lora_module_names.add(names[0] if len(names) == 1 else names[-1])

    if "lm_head" in lora_module_names:  # needed for 16-bit
        lora_module_names.remove("lm_head")
    return list(lora_module_names)


def get_max_memory_and_device_map(max_memory_mb: int):
    n_gpus = torch.cuda.device_count()
    max_memory = f"{max_memory_mb}MB"
    max_memory = {i: max_memory for i in range(n_gpus)}
    device_map = {"": 0}

    # if we are in a distributed setting, we need to set the device map and max memory per device
    if os.environ.get("LOCAL_RANK") is not None:
        local_rank = int(os.environ.get("LOCAL_RANK", "0"))
        device_map = {"": local_rank}
        max_memory = {"": max_memory[local_rank]}

    return max_memory, device_map



import functools
import itertools
import re
from collections import defaultdict
from dataclasses import dataclass, field
from functools import lru_cache
from operator import itemgetter
from typing import Any, Dict, List, Optional, Sequence

import torch
from tokenizers import Encoding
from torch.nn.functional import log_softmax
from torch.nn.utils.rnn import pad_sequence
from tqdm import trange
from transformers import (
    AutoTokenizer,
    BatchEncoding,
    GenerationConfig,
    PreTrainedModel,
    PreTrainedTokenizer,
    StoppingCriteria,
)
from transformers.generation.utils import GenerateOutput

from ykutil.log_util import log
from ykutil.python import list_squeeze

try:
    from transformer_heads.output import HeadedModelOutput

    th_available = True
except ImportError:
    th_available = False


def generate_different_sequences(
    model: PreTrainedModel,
    context: torch.Tensor,
    sequence_bias_add: int,
    sequence_bias_decay: float,
    generation_args: GenerationConfig,
    num_generations: int,
):
    gen_sequences = []
    sequence_bias = defaultdict(float)
    for _ in range(num_generations):
        gen = model.generate(
            context.unsqueeze(0),
            generation_config=generation_args,
            sequence_bias=sequence_bias or None,
        )
        if isinstance(gen, GenerateOutput):
            gen = gen.sequences
        gen = gen[0][context.shape[0] :]
        gen_sequences.append(gen)
        for key in sequence_bias:
            sequence_bias[key] *= sequence_bias_decay
        if sequence_bias_add != 0:
            for tok in gen:
                sequence_bias[tok.item()] += sequence_bias_add
    return gen_sequences


@lru_cache()
def find_tokens_with_str(tokenizer: PreTrainedTokenizer, string: str):
    return tuple(
        id
        for id in trange(len(tokenizer), desc="Find tks")
        if string in tokenizer.decode(id)
    )


def load_tk_with_pad_tk(model_path) -> PreTrainedTokenizer:
    tk = AutoTokenizer.from_pretrained(model_path)
    if tk.pad_token_id is None:
        tk.pad_token = tk.eos_token
    return tk


have_warned = False


def obtain_offsets(
    be: BatchEncoding, str_lengths: Optional[List[int]] = None
) -> list[list[tuple[int, int]]]:
    """
    >>> from transformers import AutoTokenizer
    >>> obtain_offsets(AutoTokenizer.from_pretrained("gpt2")("Hello, my dog is cute."))
    [[(0, 5), (5, 6), (6, 9), (9, 13), (13, 16), (16, 21), (21, 22)]]
    >>> tk = AutoTokenizer.from_pretrained("meta-llama/Meta-Llama-3-8B-Instruct")
    >>> tk.pad_token = tk.eos_token
    >>> obtain_offsets(tk("Hello, my dog is cute."), [22])
    Broken offsets detected. This is a known bug for llama3. Will try to fix it.
    [[(0, 0), (0, 5), (5, 6), (6, 9), (9, 13), (13, 16), (16, 21), (21, 22)]]
    >>> obtain_offsets(tk(["Hello, my dog is cute.", "Big Tree"], return_tensors="pt", padding=True), [22, 8])[1]
    [(0, 0), (0, 3), (3, 8), (0, 0), (0, 0), (0, 0), (0, 0), (0, 0)]
    """
    global have_warned
    offsets = [x.offsets for x in be.encodings]
    if not all(
        functools.reduce(
            lambda x, y: ((x[1] == y[0] or y[0] == y[1] == 0) and x[0], y[1]),
            li,
            (True, 0),
        )[0]
        for li in offsets
    ):
        if not have_warned:
            print(
                "Broken offsets detected. This is a known bug for llama3. Will try to fix it."
            )
            have_warned = True
        assert str_lengths is not None, "Cannot fix offsets without str_lengths"
        for o, sl in zip(offsets, str_lengths):
            beyond_start = False
            for i in range(len(o)):
                if o[i] != (0, 0):
                    beyond_start = True
                if i == len(o) - 1 or (beyond_start and o[i + 1] == (0, 0)):
                    if o[i] != (0, 0):
                        o[i] = (o[i][0], sl)
                else:
                    o[i] = (o[i][0], o[i + 1][0])
    return offsets


def batch_tokenization(
    tk: PreTrainedTokenizer,
    texts: List[str],
    batch_size: int,
    include_offsets: bool = False,
    **tk_args,
):
    """Use when batch of texts too large to be tokenized at once"""
    out = {"input_ids": [], "attention_mask": []}
    if include_offsets:
        out["offsets"] = []
    for s in trange(0, len(texts), batch_size, desc="Batch tokenization"):
        stuff = tk(texts[s : s + batch_size], **tk_args)
        out["input_ids"].extend(stuff["input_ids"])
        out["attention_mask"].extend(stuff["attention_mask"])
        if include_offsets:
            out["offsets"].extend(
                obtain_offsets(stuff, [len(x) for x in texts[s : s + batch_size]])
            )
    return out


def transform_with_offsets(
    offsets: list[tuple[int, int]],
    spans: list[tuple[int, int]],
    include_left=True,
    include_right=True,
):
    """
    >>> transform_with_offsets([(0, 3), (3, 6), (6, 9)], [(1, 2), (4, 8)])
    [(0, 0), (1, 2)]
    >>> transform_with_offsets([(0, 3), (3, 6), (6, 9)], [(1, 4), (5, 8)], include_left=False)
    [(1, 1), (2, 2)]
    >>> transform_with_offsets([(0, 3), (3, 6), (6, 9)], [(1, 7)], include_left=False, include_right=False)
    [(1, 1)]
    >>> transform_with_offsets([(0, 3), (3, 6), (6, 9)], [(0, 8)], include_left=False, include_right=False)
    [(0, 2)]
    """
    out_offsets = []
    offset_index = 0
    for i, num in enumerate(itertools.chain.from_iterable(spans)):
        for j in range(offset_index, len(offsets)):
            if offsets[j][0] <= num < offsets[j][1]:
                if not i % 2 and not include_left and num != offsets[j][0]:
                    out_offsets.append(j + 1)
                elif i % 2 and not include_right and num != offsets[j][1] - 1:
                    out_offsets.append(j - 1)
                else:
                    out_offsets.append(j)
                offset_index = j
                break
        else:
            assert i % 2, f"Span out of bounds, {(i,offsets[j],num)}"
            out_offsets.append(j + 1)

    return list(zip(out_offsets[::2], out_offsets[1::2]))


def regex_tokens_using_offsets(
    offsets: list[tuple[int, int]],
    text: str,
    regex: str | re.Pattern,
    include_left=True,
    include_right=True,
):
    """
    >>> regex_tokens_using_offsets([(0, 3), (3, 6), (6, 9)], "abc def ghi", "d")
    [(1, 1)]
    >>> regex_tokens_using_offsets([(0, 5), (5, 11), (11, 17)], "I am I and great", r"I \\w+",)
    [(0, 0), (1, 1)]
    """
    if isinstance(regex, str):
        regex = re.compile(regex)

    spans = [x.span() for x in regex.finditer(text)]
    return transform_with_offsets(
        offsets, spans, include_left=include_left, include_right=include_right
    )


def tokenize_instances(
    tokenizer: PreTrainedTokenizer, instances: Sequence[Dict[str, str]]
) -> Dict[str, torch.Tensor]:
    texts = [
        f"{tokenizer.bos_token}{example['text']}{tokenizer.eos_token}"
        for example in instances
    ]

    print("tokenizing all text")

    tokenized_texts = batch_tokenization(
        tk=tokenizer,
        texts=texts,
        batch_size=500,
        add_special_tokens=False,
    )
    return tokenized_texts


class TokenStoppingCriteria(StoppingCriteria):
    def __init__(self, delimiter_token):
        self.delimiter_token = (
            [delimiter_token] if type(delimiter_token) == int else delimiter_token
        )
        super().__init__()

    def __call__(self, input_ids, scores, **kwargs):
        if input_ids[0][-1] in self.delimiter_token:
            return True  # Stop generation
        return False  # Continue generation

    def __len__(self):
        return 1

    def __iter__(self):
        yield self


def flat_encode(tokenizer, inputs, add_special_tokens=False):
    if isinstance(inputs, list):
        inputs = list_squeeze(inputs)
    if isinstance(inputs, str):
        inputs = [inputs]
    if len(inputs) == 0:
        return []
    return sum(
        [tokenizer.encode(x, add_special_tokens=add_special_tokens) for x in inputs], []
    )


def get_token_begins(encoding: Encoding | BatchEncoding):
    get_start = lambda x: list(map(itemgetter(0), x))
    if isinstance(encoding, Encoding):
        return get_start(encoding.offsets)
    else:
        return [get_start(x.offsets) for x in encoding.encodings]


@dataclass
class DataCollatorWithPadding:
    """
    A data collator that pads sequences to the same length.

    Attributes:
        feature_name_to_padding_value (dict[str, int]): A dictionary mapping feature names to their padding values.

    Methods:
        __call__(features: List[Dict[str, Any]]) -> Dict[str, Any]: Pad the sequences in the features to the same length.
    """

    feature_name_to_padding_value: dict[str, int | float]
    feature_name_to_new_data_type: dict[str, torch.dtype] = field(default_factory=dict)
    padding_side: str = "right"

    def __call__(self, features: List[Dict[str, Any]]) -> Dict[str, Any]:
        """
        Pad the sequences in the features to the same length.

        Args:
            features (List[Dict[str, Any]]): A list of features, where each feature is a dictionary mapping feature names to sequences.

        Returns:
            Dict[str, Any]: A dictionary mapping feature names to padded sequences.
        """
        assert len(features) > 0, "features must not be empty."
        batch = dict()
        for key, value in self.feature_name_to_padding_value.items():
            if key in features[0]:
                if self.padding_side == "right":
                    if key in self.feature_name_to_new_data_type:
                        to_pad = [
                            feature[key]
                            .clone()
                            .detach()
                            .type(self.feature_name_to_new_data_type[key])
                            for feature in features
                        ]
                    else:
                        to_pad = [feature[key].clone().detach() for feature in features]
                    batch[key] = pad_sequence(
                        to_pad,
                        batch_first=True,
                        padding_value=value,
                    )
                elif self.padding_side == "left":
                    if key in self.feature_name_to_new_data_type:
                        to_pad = [
                            feature[key]
                            .clone()
                            .detach()
                            .flip(dims=[0])
                            .type(self.feature_name_to_new_data_type[key])
                            for feature in features
                        ]
                    else:
                        to_pad = [
                            feature[key].clone().detach().flip(dims=[0])
                            for feature in features
                        ]
                    batch[key] = pad_sequence(
                        to_pad,
                        batch_first=True,
                        padding_value=value,
                    ).flip(dims=[1])
                else:
                    raise ValueError(
                        f"padding_side must be either 'right' or 'left', but got {self.padding_side}."
                    )

        for key in features[0].keys():
            if key not in self.feature_name_to_padding_value:
                if isinstance(features[0][key], torch.Tensor):
                    try:
                        batch[key] = torch.stack(
                            [feature[key].clone().detach() for feature in features]
                        )
                    except RuntimeError as e:
                        log("Failure for key", key)
                        log(e)
                elif isinstance(features[0][key], list) and type(
                    features[0][key][0]
                ) in (
                    int,
                    float,
                ):
                    try:
                        batch[key] = torch.stack([feature[key] for feature in features])
                    except RuntimeError as e:
                        log("Failure for key", key)
                        log(e)
                else:
                    batch[key] = [feature[key] for feature in features]
                if key in self.feature_name_to_new_data_type:
                    batch[key] = batch[key].to(self.feature_name_to_new_data_type[key])

        return batch


def dict_from_chat_template(chat_template_str: str, tk_type="llama3"):
    """
    >>> from transformers import AutoTokenizer
    >>> tk = AutoTokenizer.from_pretrained("meta-llama/Meta-Llama-3-8B-Instruct")
    >>> msg_dict = [{"role":"system", "content":"You are assistant"}, {"role":"user", "content":"I am user"}, {"role":"assistant", "content":"I am assistant"}]
    >>> assert dict_from_chat_template(tk.apply_chat_template(msg_dict, tokenize=False, add_generation_prompt=False)) == msg_dict
    """
    if tk_type == "llama3":
        rex = re.compile(
            r"(<\|eot_id\|>|<\|begin_of_text\|>)?<\|start_header_id\|>(\w+)<\|end_header_id\|>\n\n(.*?)(<\|eot_id\|>|$)",
            re.DOTALL,
        )
        return [
            {"role": m.group(2), "content": m.group(3)}
            for m in rex.finditer(chat_template_str)
        ]
    elif tk_type == "gpt2":
        raise NotImplementedError("gpt2 not implemented")
    else:
        raise ValueError(f"Unknown tk_type: {tk_type}")


def tokenize(tk_name: str, text: str, add_special_tokens: bool = False):
    text = text.replace("\\n", "\n").replace("\\t", "\t")
    tk = AutoTokenizer.from_pretrained(tk_name)
    return tk(text, add_special_tokens=add_special_tokens)


def untokenize(tk_name: str, tokens: List[int]):
    tk = AutoTokenizer.from_pretrained(tk_name)
    return tk.decode(tokens)


@torch.inference_mode()
def compute_seq_log_probability(
    model: PreTrainedModel,
    pre_seq_tokens: list[int] | torch.Tensor,
    post_seq_tokens: list[int] | torch.Tensor,
    reduction: str = "sum",
) -> float:
    if isinstance(pre_seq_tokens, list):
        inputs = torch.tensor(
            pre_seq_tokens + post_seq_tokens, device=model.device
        ).unsqueeze(0)
    else:
        inputs = (
            torch.cat([pre_seq_tokens.view(-1), post_seq_tokens.view(-1)], dim=0)
            .unsqueeze(0)
            .to(model.device)
        )

    output = model(inputs)
    if th_available and isinstance(output, HeadedModelOutput):
        logits = output.preds_by_head["lm_head"][
            0, -len(post_seq_tokens) - 1 : -1
        ].cpu()
    else:
        logits = output.logits[0, -len(post_seq_tokens) - 1 : -1].cpu()

    logprobs = log_softmax(logits, dim=-1)

    red = torch.mean if reduction == "mean" else torch.sum

    res = float(
        red(logprobs.gather(1, torch.tensor(post_seq_tokens).unsqueeze(1)).squeeze())
    )

    return res


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import argparse
import os
import json
from ast import literal_eval

from fire import Fire
from importlib.util import find_spec

if find_spec("pandas") is not None:
    from ykutil.pandas_util import df_to_csv
if (
    find_spec("datasets") is not None
    and find_spec("transformers") is not None
    and find_spec("torch") is not None
):
    from ykutil.dataset import colorcode_entry, count_token_sequence
    from ykutil.dataset import describe_dataset as ds_describe_dataset
from ykutil.tools import bulk_rename
from ykutil.sql import merge_databases


def do_bulk_rename():
    Fire(bulk_rename)


def beautify_json(json_str: str):
    print(json.dumps(json.loads(json_str), indent=4))


def describe_dataset(
    ds_name: str,
    tokenizer_name: str = None,
    show_rows: tuple = (0,),
    zip_labels=False,
):
    from datasets import Dataset
    from transformers import AutoTokenizer

    ds = Dataset.load_from_disk(ds_name)
    tokenizer = (
        AutoTokenizer.from_pretrained(tokenizer_name) if tokenizer_name else None
    )
    return ds_describe_dataset(ds, tokenizer, show_rows, zip_labels)


def tokenize(tk: str, text: str, add_special_tokens: bool = False):
    from ykutil.transformer import tokenize as tk_tokenize

    return tk_tokenize(tk_name=tk, text=text, add_special_tokens=add_special_tokens)


def untokenize(tk: str, tokens: str):
    from ykutil.transformer import untokenize as tk_untokenize

    return tk_untokenize(tk_name=tk, tokens=literal_eval(tokens))


def do_merge_databases():
    parser = argparse.ArgumentParser()
    parser.add_argument("target_db", type=str)
    parser.add_argument("source_dbs", type=str, nargs="+")
    args = parser.parse_args()
    return merge_databases(args.target_db, args.source_dbs)


def do_merge_database_folder():
    parser = argparse.ArgumentParser()
    parser.add_argument("target_db", type=str)
    parser.add_argument("source_folder", type=str)
    args = parser.parse_args()

    source_dbs = [
        os.path.join(args.source_folder, f) for f in os.listdir(args.source_folder)
    ]

    return merge_databases(args.target_db, source_dbs)


def do_describe_dataset():
    parser = argparse.ArgumentParser()
    parser.add_argument("ds_name", type=str)
    parser.add_argument("--tokenizer_name", type=str, default=None)
    parser.add_argument("--show_rows", type=int, nargs="+", default=(0,))
    parser.add_argument("--zip_labels", action="store_true")
    args = parser.parse_args()
    return describe_dataset(
        ds_name=args.ds_name,
        tokenizer_name=args.tokenizer_name,
        show_rows=tuple(args.show_rows),
        zip_labels=args.zip_labels,
    )


def do_tokenize():
    parser = argparse.ArgumentParser()
    parser.add_argument("tk", type=str)
    parser.add_argument("text", type=str)
    parser.add_argument("--add_special_tokens", action="store_true")
    args = parser.parse_args()
    return tokenize(
        tk=args.tk, text=args.text, add_special_tokens=args.add_special_tokens
    )


def do_untokenize():
    parser = argparse.ArgumentParser()
    parser.add_argument("tk", type=str)
    parser.add_argument("tokens", type=str)
    args = parser.parse_args()
    return untokenize(tk=args.tk, tokens=args.tokens)


def do_colorcode_dataset():
    parser = argparse.ArgumentParser()
    parser.add_argument("ds_path", type=str)
    parser.add_argument("--tokenizer", type=str, default="mistralai/Mistral-7B-v0.1")
    parser.add_argument("--output", type=str, default=None)
    parser.add_argument("--start", type=int, default=0)
    parser.add_argument("--end", type=int, default=1)
    parser.add_argument("--ugly", action="store_true")
    args = parser.parse_args()
    return colorcode_entry(
        token_ds_path=args.ds_path,
        fname=args.output,
        tokenizer_path=args.tokenizer,
        num_start=args.start,
        num_end=args.end,
        beautify=not args.ugly,
    )


def do_count_token_sequence():
    parser = argparse.ArgumentParser()
    parser.add_argument("ds_path", type=str)
    parser.add_argument("token_sequence", type=str)
    parser.add_argument("--start", type=int, default=0)
    parser.add_argument("--end", type=int, default=1)
    args = parser.parse_args()
    token_sequence = literal_eval(args.token_sequence)

    count = count_token_sequence(
        token_ds_path=args.ds_path,
        token_sequence=token_sequence,
        num_start=args.start,
        num_end=args.end,
    )
    print(count)


def do_beautify_json():
    parser = argparse.ArgumentParser()
    parser.add_argument("json_str", type=str)
    args = parser.parse_args()
    return beautify_json(json_str=args.json_str)


def do_df_to_csv():
    parser = argparse.ArgumentParser()
    parser.add_argument("filename", type=str)
    args = parser.parse_args()
    return df_to_csv(args.filename)



import numpy as np
import torch
from torch.nn.functional import sigmoid
from transformer_heads.model.model import HeadedModel
from transformer_heads.output import HeadedModelOutput
from transformers import TrainerCallback
from transformers.modeling_outputs import CausalLMOutputWithPast

from ykutil.constants import IGNORE_INDEX


@torch.inference_mode()
def compute_policy_metrics(
    outputs: CausalLMOutputWithPast,
    input_dic: dict[str, torch.Tensor],
    model: torch.nn.Module,
    **kwargs,
):
    labels = input_dic["labels"]
    metrics = {}
    distrib = torch.softmax(outputs.logits.detach(), dim=-1)
    correctness = (distrib[:, :-1].argmax(-1) == labels[:, 1:]).float()
    correctness = correctness[labels[:, 1:] != IGNORE_INDEX]
    metrics["policy_accuracy"] = correctness.mean().item()
    metrics["policy_entropy"] = -torch.mean(
        torch.sum(distrib * torch.log(distrib), dim=-1)
    ).item()
    return metrics


@torch.inference_mode()
def compute_value_metrics(
    outputs: HeadedModelOutput,
    input_dic: dict[str, torch.Tensor],
    model: HeadedModel,
    head_remap: dict[str, str] = {},
):
    metrics = {}
    for key, value in outputs.loss_by_head.items():
        metrics[f"loss_{key}"] = float(value)
        if (
            outputs.adapted_loss_by_head is not None
            and key in outputs.adapted_loss_by_head
        ):
            adapted = outputs.adapted_loss_by_head[key]
            if adapted != value:
                metrics[f"adapted_loss_{key}"] = float(adapted)

    for head_name, preds in outputs.preds_by_head.items():
        if head_name == "lm_head":
            continue
        if head_name in head_remap:
            labs = input_dic[head_remap[head_name]]
        else:
            labs = input_dic[model.head_configs[head_name].target]

        if "logits" in model.head_configs[head_name].loss_fct:
            preds = sigmoid(preds[labs != IGNORE_INDEX]).detach()
        else:
            preds = preds[labs != IGNORE_INDEX].detach()
        labs = labs[labs != IGNORE_INDEX]

        mse = torch.mean((preds - labs) ** 2)

        metrics[f"{head_name}_pred_avg"] = preds.mean().item()
        metrics[f"{head_name}_pred_std"] = preds.std().item()
        metrics[f"{head_name}_label_avg"] = labs.float().mean().item()
        metrics[f"{head_name}_mse"] = mse.item()
        metrics[f"{head_name}_mae"] = torch.abs(preds - labs).mean().item()

    return metrics


@torch.inference_mode()
def compute_classification_head_metrics(
    outputs: HeadedModelOutput,
    input_dic: dict[str, torch.Tensor],
    model: HeadedModel,
    head_remap: dict[str, str] = {},
):
    metrics = {}

    for key, value in outputs.loss_by_head.items():
        metrics[f"loss_{key}"] = float(value)
        if (
            outputs.adapted_loss_by_head is not None
            and key in outputs.adapted_loss_by_head
        ):
            adapted = outputs.adapted_loss_by_head[key]
            if adapted != value:
                metrics[f"adapted_loss_{key}"] = float(adapted)

    if "lm_head" in outputs.preds_by_head:
        preds = outputs.preds_by_head["lm_head"].detach().argmax(-1)
        labels = input_dic["labels"]
        correctness = (preds[:, :-1] == labels[:, 1:]).float()
        correctness = correctness[labels[:, 1:] != IGNORE_INDEX]
        metrics["policy_accuracy"] = correctness.mean().item()

    for head_name, preds in outputs.preds_by_head.items():
        if head_name == "lm_head":
            continue
        if head_name in head_remap:
            labs = input_dic[head_remap[head_name]]
        else:
            labs = input_dic[model.head_configs[head_name].target]

        if "logits" in model.head_configs[head_name].loss_fct:
            preds = sigmoid(preds[labs != IGNORE_INDEX]).detach()
        else:
            preds = preds[labs != IGNORE_INDEX].detach()
        labs = labs[labs != IGNORE_INDEX]

        assert torch.max(preds) <= 1.0 and torch.min(preds) >= 0.0

        preds_round = torch.round(preds)
        correctness = (preds_round == labs).float()
        mse = torch.mean((preds - labs) ** 2)

        # true_positives = (preds_round * labs).sum()
        # false_positives = (preds_round * (1 - labs)).sum()
        # false_negatives = ((1 - preds_round) * labs).sum()
        # metrics[f"{head_name}_true_positives"] = true_positives.item()
        # metrics[f"{head_name}_false_positives"] = false_positives.item()
        # metrics[f"{head_name}_false_negatives"] = false_negatives.item()

        # precision = true_positives / (true_positives + false_positives)
        # recall = true_positives / (true_positives + false_negatives)

        precision = (preds_round * labs).sum() / preds_round.sum()
        recall = (preds_round * labs).sum() / labs.sum()

        f1 = 2 * precision * recall / (precision + recall)
        metrics[f"{head_name}_first_pred"] = preds[0].item()
        metrics[f"{head_name}_accuracy"] = correctness.mean().item()
        metrics[f"{head_name}_bias"] = preds.mean().item()
        metrics[f"{head_name}_std"] = preds.std().item()
        metrics[f"{head_name}_precision"] = precision.item()
        metrics[f"{head_name}_recall"] = recall.item()
        metrics[f"{head_name}_f1"] = f1.item()
        metrics[f"{head_name}_label_avg"] = labs.float().mean().item()
        metrics[f"{head_name}_mse"] = mse.item()

    return metrics


class EvaluateFirstStepCallback(TrainerCallback):
    def on_step_begin(self, args, state, control, **kwargs):
        if state.global_step == 0 and args.logging_first_step:
            control.should_evaluate = True


compute_metrics_functions = dict(
    compute_classification_head_metrics=compute_classification_head_metrics,
    compute_policy_metrics=compute_policy_metrics,
    compute_value_metrics=compute_value_metrics,
)



import pandas as pd
import pandera as pa


def empty_dataframe_from_model(Model: pa.DataFrameModel) -> pd.DataFrame:
    schema = Model.to_schema()
    return pd.DataFrame(columns=schema.dtypes.keys()).astype(
        {col: str(dtype) for col, dtype in schema.dtypes.items()}
    )



import random
import math
import re
from itertools import groupby
from typing import Any, Iterable, List, Optional

from tqdm import tqdm

from ykutil.types_util import SimpleGenerator, T, U


def identity(x):
    return x


def dictsum(*dicts: list[dict[Any, int | float]]) -> dict[Any, int | float]:
    """
    >>> dictsum({"a": 1, "b": 2}, {"a": 3, "b": 4})
    {'a': 4, 'b': 6}
    """
    out = {}
    for d in dicts:
        for k, v in d.items():
            out[k] = out.get(k, 0) + v
    return out


def first_true(iterable):
    for i, x in enumerate(iterable):
        if x:
            return i


def transpose_list(lst: list[list]) -> list[list]:
    """
    >>> transpose_list([[1, 2, 3], [4, 5, 6]])
    [[1, 4], [2, 5], [3, 6]]
    """
    return [list(x) for x in zip(*lst)]


def list_get(lst: list, indices: list) -> list:
    """
    >>> list_get([1, 2, 3, 4, 5], [0, 2, 3])
    [1, 3, 4]
    """
    return [lst[i] for i in indices]


def vector_sum(list_a, list_b):
    """
    >>> vector_sum([1, 2, 3], [4, 5, 6])
    [5, 7, 9]
    """
    return [a + b for a, b in zip(list_a, list_b)]


def list_rindex(li, x):
    """
    >>> list_rindex([1, 2, 3, 1, 2, 3, 4], 3)
    5
    """
    for i in reversed(range(len(li))):
        if li[i] == x:
            return i
    raise ValueError(f"{x} is not in list")


def list_split(li: list, max_len: int, min_len: Optional[int] = None) -> list[list]:
    """
    >>> list_split([1, 2, 3, 4, 5, 1, 2, 3], 3)
    [[1, 2, 3], [4, 5, 1], [2, 3]]
    """
    return [
        li[i : i + max_len]
        for i in range(0, len(li), max_len)
        if min_len is None or len(li[i : i + max_len]) >= min_len
    ]


def list_split_at_value(li: list, value: Any) -> list[list]:
    """
    >>> list_split_at_value([1, 2, 3, 4, 5, 1, 2, 3], 1)
    [[], [2, 3, 4, 5], [2, 3]]
    """
    splits = []
    start = 0
    for i, x in enumerate(li):
        if x == value:
            splits.append(li[start:i])
            start = i + 1
    splits.append(li[start:])
    return splits


def shortest_common_supersequence(lists: list[list]) -> list:
    """Find the shortest common supersequence that contains
    the given lists as subsequences. All elements appear in
    an order that respects the relative orderings of each list.


    >>> shortest_common_supersequence([[1,7,2,3], [1,2,8,9,3], [5,1,2,3,6]])
    [5, 1, 7, 2, 8, 9, 3, 6]
    >>> shortest_common_supersequence([[1,2,3,3,2,1], [2,3,3,4,2,1]])
    [1, 2, 3, 3, 4, 2, 1]
    """
    # Credits to o1

    # If there's only one list, that list itself is the shortest common supersequence.
    if len(lists) == 1:
        return lists[0]

    def scs_two(a: list, b: list) -> list:
        # Compute the shortest common supersequence of two lists a and b.

        # First, compute the LCS (Longest Common Subsequence) DP table.
        len_a, len_b = len(a), len(b)
        dp = [[0] * (len_b + 1) for _ in range(len_a + 1)]

        for i in range(1, len_a + 1):
            for j in range(1, len_b + 1):
                if a[i - 1] == b[j - 1]:
                    dp[i][j] = dp[i - 1][j - 1] + 1
                else:
                    dp[i][j] = max(dp[i - 1][j], dp[i][j - 1])

        # Reconstruct the shortest common supersequence from the DP table.
        i, j = len_a, len_b
        scs = []
        while i > 0 and j > 0:
            if a[i - 1] == b[j - 1]:
                # If elements are equal, it must be in the SCS.
                scs.append(a[i - 1])
                i -= 1
                j -= 1
            else:
                # Pick the direction which gives us the longer LCS and add that element.
                if dp[i - 1][j] > dp[i][j - 1]:
                    scs.append(a[i - 1])
                    i -= 1
                else:
                    scs.append(b[j - 1])
                    j -= 1

        # If any elements are left in a or b, append them.
        while i > 0:
            scs.append(a[i - 1])
            i -= 1
        while j > 0:
            scs.append(b[j - 1])
            j -= 1

        scs.reverse()
        return scs

    # Iteratively combine all lists into a single shortest common supersequence.
    current_scs = lists[0]
    for i in range(1, len(lists)):
        current_scs = scs_two(current_scs, lists[i])

    return current_scs


def split_multi(
    lists: list[list], max_len: int, progress=False, min_len: Optional[int] = None
) -> tuple[list[list], list]:
    """
    Slow for large lists
    >>> split_multi([[1, 2, 3, 4], [1, 2], [3, 4, 5, 6, 7]], 2)
    ([[1, 2], [3, 4], [1, 2], [3, 4], [5, 6], [7]], [0, 0, 1, 2, 2, 2])
    """
    splits = [
        list_split(x, max_len, min_len=min_len)
        for x in tqdm(lists, disable=not progress, desc="performing list splits")
    ]
    return sum(splits, []), sum(
        [
            [i] * len(x)
            for i, x in tqdm(
                enumerate(splits),
                total=len(splits),
                disable=not progress,
                desc="concatenating splits",
            )
        ],
        [],
    )


def list_multiply(elem_list: list, mul_list: list) -> list:
    """
    >>> list_multiply([3, 4, 3], [1, 3, 2])
    [3, 4, 4, 4, 3, 3]
    """
    return sum([[e] * m for e, m in zip(elem_list, mul_list)], [])


def index_of_sublist_match(haystack: List[list], needle: list) -> int:
    """
    >>> index_of_sublist_match([[3, 4, 3], [1, 3, 2, 4], [1, 2]], [1, 3])
    1
    """
    return [x[: len(needle)] for x in haystack].index(needle)


def count_sublist_matches(haystack: List[list], needle: list) -> int:
    """
    >>> count_sublist_matches([[3, 4, 3], [1, 3, 2, 4], [1, 2]], [1, 3])
    1
    """
    count = 0
    for sublist in haystack:
        if sublist[: len(needle)] == needle:
            count += 1
    return count


def anyin(needles: Iterable, haystack: Iterable) -> bool:
    """
    >>> anyin([1, 2, 3], [3, 4, 5])
    True
    """
    return any(x in haystack for x in needles)


def list_squeeze(l: list) -> list:
    """
    >>> list_squeeze([[1], [2], [3], [[3, 4, 5]]])
    [1, 2, 3, [3, 4, 5]]
    """
    if isinstance(l, list):
        if len(l) == 1:
            return list_squeeze(l[0])
        elif len(l) == 0:
            return l
        else:
            return [list_squeeze(x) for x in l]
    else:
        return l


def multify_text(text: str, roles: list[str | None]) -> list[str]:
    """
    >>> multify_text("Player {pid} ({role}) eats a cake", ["fascist", None, "liberal"])
    ['Player 1 (fascist) eats a cake', 'Player 2', 'Player 3 (liberal) eats a cake']
    """
    return [
        (
            text.split(" ({role})")[0].format(pid=i + 1)
            if r is None
            else text.format(pid=i + 1, role=r)
        )
        for i, r in enumerate(roles)
    ]


def naive_regex_escape(some_str: str) -> str:
    # This method should be already implemented in some standard python module,
    # but I can not find it. (re.escape does not seem to do what I want)
    r"""
    >>> naive_regex_escape(r"Peter ({Person}) is eating [3 or 4] sandwiches.")
    'Peter \\({Person}\\) is eating \\[3 or 4\\] sandwiches\\.'
    """
    # This metachar map sounds stupid, but python syntax is weird here
    metachar_map = {
        r"(": r"\(",
        r")": r"\)",
        r"]": r"\]",
        r"[": r"\[",
        r"^": r"\^",
        r"$": r"\$",
        r"+": r"\+",
        r"*": r"\*",
        r".": r"\.",
        r"?": r"\?",
    }
    for key, value in metachar_map.items():
        some_str = some_str.replace(key, value)
    return some_str


def chunk_list(lst, n) -> SimpleGenerator[list]:
    """Returns a generator that yields consecutive chunks of size n from lst.
    Works with huggingface datasets as well
    """
    for i in range(0, len(lst), n):
        yield [
            lst[x] for x in range(i, min(i + n, len(lst)))
        ]  # Sounds stupid, but so are huggingface datasets


def list_flip(lst: list[int | float]) -> list[int | float]:
    """
    >>> list_flip([1, 2, 4, 3])
    [4, 3, 1, 2]
    """
    mx = max(lst)
    mn = min(lst)
    new_lst = [mx + mn - x for x in lst]
    return new_lst


def approx_number_split(n: int, n_splits: int) -> list[int]:
    """
    >>> approx_number_split(10, 3)
    [4, 3, 3]
    """
    out = [n // n_splits] * n_splits
    for i in range(n % n_splits):
        out[i] += 1
    return out


def approx_list_split(lst: list, n_splits: int) -> SimpleGenerator[list]:
    """
    >>> list(approx_list_split([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], 3))
    [[0, 1, 2, 3], [4, 5, 6], [7, 8, 9]]
    """
    splits = approx_number_split(len(lst), n_splits)

    start = 0
    for sp in splits:
        yield lst[start : start + sp]
        start += sp


def list_find_new(list_orig: list, list_new: list) -> list:
    """list_new contains more content at the end than list_orig
    but is also missing some parts of list_orig. This function
    will find the new content in list_new and return it.

    >>> list_find_new([1, 2, 3, 4], [1, 3, 4, 5, 6])
    [5, 6]
    """
    # Credits to o1
    last_match_index = -1  # position of the last matched element in list_new
    for elem in list_orig:
        # Find elem in list_new starting from last_match_index+1
        try:
            pos = list_new.index(elem, last_match_index + 1)
            last_match_index = pos
        except ValueError:
            # elem is not found in list_new after last_match_index, skip it
            continue

    # Everything after last_match_index in list_new is new
    new_part = list_new[last_match_index + 1 :]
    return new_part


def check_if_in_other_list(lst: list, other: list) -> list[bool]:
    """other may be shorter than lst and may be missing parts of
    lst inbetween. This function will find all overlaps between
    the two lists and return a list of booleans indicating which
    parts of lst are in other.

    The logic is that `other` should be considered as a subsequence
    to be matched in order. We iterate through `lst` and try to match
    elements of `other` in sequence. When we find a match, we move on
    to the next element of `other`. If there's no match, we mark False.

    >>> check_if_in_other_list([1, 6, 3, 4, 5, 6], [1, 3, 4, 6])
    [True, False, True, True, False, True]
    """
    # Credits to o1
    result = []
    other_idx = 0
    other_len = len(other)

    for elem in lst:
        if other_idx < other_len and elem == other[other_idx]:
            # Matches the next required element in 'other'
            result.append(True)
            other_idx += 1
        else:
            # Does not match the next required element
            result.append(False)

    return result


def nth_index(lst: list, value, n: int) -> int:
    """
    >>> nth_index([1, 2, 3, 1, 2, 3, 4], 3, 2)
    5
    """
    start = -1
    for _ in range(n):
        start = lst.index(value, start + 1)
    return start


def update_running_avg(
    old_avg: float, old_weight: int | float, new_avg: float, new_weight=1
) -> float:
    """
    >>> update_running_avg(0, 9, 1, 1)
    0.1
    """
    return (old_avg * old_weight + new_avg * new_weight) / (old_weight + new_weight)


def transpose_li_of_dict(lidic: list[dict]) -> dict[Any, list[Any]]:
    """
    >>> transpose_li_of_dict([{"a": 1, "b": 2}, {"a": 3, "b": 4}])
    {'a': [1, 3], 'b': [2, 4]}
    """
    return {k: [d[k] for d in lidic] for k in lidic[0].keys()}


def transpose_dict_of_li(d: dict[Any, list]) -> list[dict[Any, Any]]:
    """
    >>> transpose_dict_of_li({"a": [1, 3], "b": [2, 4]})
    [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}]
    """
    return [dict(zip(d.keys(), x)) for x in zip(*d.values())]


def flatten(li: list[list]) -> list:
    """
    >>> flatten([[1, 2], [3, 4]])
    [1, 2, 3, 4]
    """
    return (
        [item for sublist in li for item in sublist]
        if isinstance(li[0], Iterable)
        else li
    )


def all_equal(iterable, force_value=None) -> bool:
    """
    >>> all_equal([1, 1, 1, 1], 1)
    True
    """
    g = groupby(iterable)
    if force_value is not None:
        return next(g, (force_value,))[0] == force_value and not next(g, False)
    return next(g, True) and not next(g, False)


def str_find_all(string: str, sub: str):
    """
    >>> list(str_find_all("abcabcabc", "abc"))
    [0, 3, 6]
    """
    start = 0
    while True:
        start = string.find(sub, start)
        if start == -1:
            break
        yield start
        start += len(sub)


def re_line_matches(string: str, regex: str | re.Pattern):
    r"""
    >>> list(re_line_matches("a bus\nich muss\nyeah\neine nuss", r"[a-z]us"))
    [0, 1, 3]
    """
    if isinstance(regex, str):
        regex = re.compile(regex)
    assert isinstance(regex, re.Pattern)
    newlines = str_find_all(string, "\n")
    matches = regex.finditer(string)

    nl_idx = 0
    nl_pos = 0
    for m in matches:
        if m.span()[1] <= nl_pos:  # Ensure just one match per line
            continue
        while m.span()[1] > nl_pos:
            nl_idx += 1
            try:
                nl_pos = next(newlines)
            except StopIteration:
                nl_pos = len(string)
                assert m.span()[1] <= nl_pos
        yield nl_idx - 1


def make_list_unique(seq):
    """
    >>> list(make_list_unique([1, 2, 2, 3, 3, 3, 4, 4, 4, 4]))
    [1, 2, 3, 4]
    """
    seen = set()
    for x in seq:
        if x in seen:
            continue
        seen.add(x)
        yield x


def dict_percentages(d: dict[Any, int | float]):
    """
    >>> dict_percentages({"a": 3, "b": 2, "c": 5})
    {'a': 0.3, 'b': 0.2, 'c': 0.5}
    """
    tot = sum(d.values())
    return {k: v / tot for k, v in d.items()}


def recursed_dict_percentages(d: dict):
    """
    >>> recursed_dict_percentages({"a": {"b": 3, "c": 2}, "d": {"e": 7, "f": 3}})
    {'a': {'b': 0.6, 'c': 0.4}, 'd': {'e': 0.7, 'f': 0.3}}
    """
    if len(d) == 0:
        return d
    if type(next(iter(d.values()))) in (int, float):
        return dict_percentages(d)
    return {k: recursed_dict_percentages(v) for k, v in d.items()}


def recursed_merge_percent_stats(lst: list[dict | float | int], weights=None):
    """
    >>> recursed_merge_percent_stats([{"a": {"b": 0.3, "c": 0.2}, "d": {"e": 0.4, "f": 0.3}}, {"a": {"b": 0.4, "c": 0.3}, "d": {"e": 0.5, "f": 0.4}}])
    {'a': {'b': 0.35, 'c': 0.25}, 'd': {'e': 0.45, 'f': 0.35}}
    """
    if weights is None:
        weights = [1] * len(lst)

    if isinstance(lst[0], dict):
        return {
            k: (
                recursed_merge_percent_stats(
                    [d[k] for d in lst if k in d],
                    weights=[w for w, d in zip(weights, lst) if k in d],
                )
            )
            for k in make_list_unique(sum((list(x.keys()) for x in lst), []))
        }
    else:
        return sum(x * w for x, w in zip(lst, weights)) / sum(weights)


def recursed_sum_up_stats(lst: list[dict | int]):
    """
    >>> recursed_sum_up_stats([{"a": {"b": 3, "c": 2}, "d": {"e": 7, "f": 3}}, {"a": {"b": 4, "c": 3}, "d": {"e": 5, "f": 4}}])
    {'a': {'b': 7, 'c': 5}, 'd': {'e': 12, 'f': 7}}
    """
    if isinstance(lst[0], dict):
        return {
            k: recursed_sum_up_stats([d[k] for d in lst if k in d])
            for k in make_list_unique(sum((list(x.keys()) for x in lst), []))
        }
    else:
        return sum(lst)


def removesuffixes(lst: list[T], suffix: Iterable[T]) -> list[T]:
    """
    >>> removesuffixes([1, 2, 3, 4, 3], [3,4])
    [1, 2]
    >>> removesuffixes([1, 2, 3, 4, 3], [3,2])
    [1, 2, 3, 4]
    """
    while len(lst) > 0 and lst[-1] in suffix:
        lst.pop()
    return lst


def dict_with(d: dict[str, U], **kwargs: U) -> dict[str, U]:
    """
    >>> dict_with({"a": 1, "b": 2}, b=3)
    {'a': 1, 'b': 3}
    """
    return {k: kwargs.get(k, v) for k, v in d.items()}


def dict_without(d: dict[T, U], without: str | Iterable[T]) -> dict[T, U]:
    """
    >>> dict_without({"a": 1, "b": 2, "c": 3}, "a")
    {'b': 2, 'c': 3}
    >>> dict_without({"a": 1, "b": 2, "c": 3}, ["a", "b"])
    {'c': 3}
    """
    if isinstance(without, str):
        without = [without]
    new_d = d.copy()
    for key in without:
        new_d.pop(key, None)
    return new_d


def all_sublist_matches(lst: list, sublst: list):
    """
    >>> list(all_sublist_matches([1, 2, 3, 4, 3, 2, 2, 3, 4], [2, 3]))
    [1, 6]
    """
    for i in range(len(lst) - len(sublst) + 1):
        if lst[i : i + len(sublst)] == sublst:
            yield i


def list_in_list(lst: list, sublst: list) -> bool:
    """
    >>> list_in_list([1, 2, 3, 4, 3, 2, 2, 3, 4], [2, 3])
    True
    """
    for i in range(len(lst) - len(sublst) + 1):
        if lst[i : i + len(sublst)] == sublst:
            return True
    return False


def count_sublist_occurrences(lst: list, sublst: list) -> int:
    """
    >>> count_sublist_occurrences([1, 2, 3, 4, 3, 2, 2, 3, 4], [2, 3])
    2
    """
    count = 0
    for i in range(len(lst) - len(sublst) + 1):
        if lst[i : i + len(sublst)] == sublst:
            count += 1
    return count


def remove_until(lst: list[T], elem: T, remove_elem=True) -> list[T]:
    """
    >>> remove_until([1, 2, 3, 4, 3, 2, 2, 3, 4], 3, remove_elem=False)
    [3, 4, 3, 2, 2, 3, 4]
    >>> remove_until([1, 2, 3, 4, 3, 2, 2, 3, 4], 3, remove_elem=True)
    [4, 3, 2, 2, 3, 4]
    """
    try:
        idx = lst.index(elem)
    except ValueError:
        return lst

    if remove_elem:
        return lst[idx + 1 :]
    return lst[idx:]


def unique_n_times(
    lst: list,
    n: int,
    invalid_filter: set = set(),
    verbose=False,
    comboer: Optional[list] = None,
    shuffle: bool = False,
) -> list[int]:
    """
    Returns the indices of the first n times each unique element appears in the list
    If a comboer is given, all indices with the same unique element and comboer are treated as the same element

    >>> unique_n_times([0,2,1,2,2,1,0,0,1,2], 2)
    [0, 1, 2, 3, 5, 6]
    >>> unique_n_times([0,2,1,2,2,1,0,0,1,2], 2, comboer=[5,8,4,9,10,5,5,6,6,8])
    [0, 1, 2, 3, 5, 6, 7, 9]
    """
    if shuffle:
        if comboer is not None:
            lst_and_comboer = list(zip(lst, comboer))
            random.shuffle(lst_and_comboer)
            lst, comboer = zip(*lst_and_comboer)
        else:
            lst = lst.copy()
            random.shuffle(lst)

    seen = {}
    result = []
    combos = set()
    for i, x in tqdm(
        enumerate(lst), total=len(lst), disable=not verbose, desc="unique_n_times"
    ):
        if i not in invalid_filter:
            if comboer is not None:
                if (x, comboer[i]) in combos:
                    result.append(i)
                    continue
            if x in seen:
                seen[x] += 1
            else:
                seen[x] = 1
            if seen[x] <= n:
                result.append(i)
                if comboer is not None:
                    combos.add((x, comboer[i]))
    return result


def generate_random_with_cross_sum(n: int):
    left_over = n
    digits = []
    while left_over > 9:
        digit = random.randint(0, 9)
        digits.append(digit)
        left_over -= digit
    digits.append(left_over)
    return "".join(map(str, digits))


def nanmean(values: list[float]) -> float:
    """
    >>> nanmean([1, 2, 3, 4, 5, float("nan")])
    3.0
    """
    valid_values = [v for v in values if not math.isnan(v)]
    return (
        float("nan")
        if len(valid_values) == 0
        else sum(valid_values) / len(valid_values)
    )


def weighted_nanmean(values: list[float], weights: list[float]) -> float:
    """
    >>> weighted_nanmean([1, 2, 3, 4, 5, float("nan")], [1, 1, 1, 1, 1, 0])
    3.0
    """
    valid_muls = [v * w for v, w in zip(values, weights) if not math.isnan(v)]
    valid_weights = [w for v, w in zip(values, weights) if not math.isnan(v)]
    return (
        float("nan") if len(valid_muls) == 0 else sum(valid_muls) / sum(valid_weights)
    )


def compute_gini(values: list[int | float]) -> float:
    """Compute the Gini coefficient for a list of values.

    The Gini coefficient measures inequality in a distribution, ranging from
    0 (perfect equality) to 1 (perfect inequality).

    >>> compute_gini([1, 1, 1, 1])
    0.0
    >>> compute_gini([0, 0, 0, 100])
    0.75
    >>> round(compute_gini([1, 2, 3, 4, 5]), 4)
    0.2667
    >>> compute_gini([10])
    0.0
    >>> compute_gini([])
    0.0
    """
    n = len(values)
    if n <= 1:
        return 0.0

    total = sum(values)
    if total == 0:
        return 0.0

    sorted_values = sorted(values)
    cumulative = 0.0
    for i, v in enumerate(sorted_values):
        cumulative += (i + 1) * v

    return (2 * cumulative - (n + 1) * total) / (n * total)


def levenshtein_distance(a: str | list, b: str | list) -> int:
    """Compute the Levenshtein (edit) distance between two sequences.

    The Levenshtein distance is the minimum number of single-element edits
    (insertions, deletions, or substitutions) required to change one
    sequence into another.

    >>> levenshtein_distance("kitten", "sitting")
    3
    >>> levenshtein_distance("saturday", "sunday")
    3
    >>> levenshtein_distance([1, 2, 3], [1, 3, 2, 4])
    2
    >>> levenshtein_distance("", "abc")
    3
    >>> levenshtein_distance("same", "same")
    0
    """
    len_a, len_b = len(a), len(b)

    # Use a single row for space optimization
    prev = list(range(len_b + 1))
    curr = [0] * (len_b + 1)

    for i in range(1, len_a + 1):
        curr[0] = i
        for j in range(1, len_b + 1):
            if a[i - 1] == b[j - 1]:
                curr[j] = prev[j - 1]
            else:
                curr[j] = 1 + min(prev[j], curr[j - 1], prev[j - 1])
        prev, curr = curr, prev

    return prev[len_b]


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import os


def find_all_file_paths(folder: str, force_ending=None):
    findings = []
    for root, _, files in os.walk(folder):
        for file in files:
            if force_ending is None or file.endswith(force_ending):
                findings.append(os.path.join(root, file))
    return findings


def search_file_in_folder(fname: str, folder: str):
    if os.path.isfile(fname):
        return fname
    assert os.path.isdir(folder), f"{folder} is not a folder"

    findings = []
    for root, _, files in os.walk(folder):
        if fname in files:
            findings.append(os.path.join(root, fname))

    assert len(findings) > 0, f"{fname} not found in {folder}"
    assert len(findings) == 1, f"{fname} found in multiple places: {findings}"
    return findings[0]



import base64
from mimetypes import guess_type

from ykutil import AzureModelWrapper


# Function to encode a local image into data URL
def local_image_to_data_url(image_path):
    # Guess the MIME type of the image based on the file extension
    mime_type, _ = guess_type(image_path)
    if mime_type is None:
        mime_type = "application/octet-stream"  # Default MIME type if none is found

    # Read and encode the image file
    with open(image_path, "rb") as image_file:
        base64_encoded_data = base64.b64encode(image_file.read()).decode("utf-8")

    # Construct the data URL
    return f"data:{mime_type};base64,{base64_encoded_data}"


# Example usage
image_path = "/mnt/c/Users/ykeller/Pictures/buff_llama.png"
data_url = local_image_to_data_url(image_path)
print("Data URL:", data_url)

messages = [
    {
        "role": "user",
        "content": [
            {"type": "text", "text": "What’s in this image?"},
            {
                "type": "image_url",
                "image_url": {
                    "url": data_url,
                },
            },
        ],
    }
]

model = AzureModelWrapper()

print(model.complete(messages))



IGNORE_INDEX = -100
DEFAULT_PAD_TOKEN = "[PAD]"
SPACE_TOKENIZERS = (
    "meta-llama/Meta-Llama-3-8B",
    "meta-llama/Meta-Llama-3-8B-Instruct",
    "meta-llama/Meta-Llama-3-70B",
    "meta-llama/Meta-Llama-3-70B-Instruct",
)



import logging
import time
import asyncio
from ykutil.log_util import TruncatingFileHandler, RotatingFileHandle
import os
from fire import Fire


def test_truncating_file_handler():
    fl = TruncatingFileHandler("test.log", max_size_mb=0.01, keep_rate=0.5)
    logger = logging.getLogger("test")
    logger.setLevel(logging.DEBUG)
    logger.addHandler(fl)
    for i in range(100):
        logger.info("test" * 100)
    assert os.path.exists("test.log")
    print(os.path.getsize("test.log"), fl.max_size_bytes)


async def test_rotating_file_handle():
    """Test that RotatingFileHandle limits file size when used with create_subprocess_shell"""
    log_file = "test_rotating.log"
    max_size_mb = 0.01  # Very small limit for testing

    # Clean up any existing test file
    if os.path.exists(log_file):
        os.remove(log_file)

    # Create RotatingFileHandle
    log_handle = RotatingFileHandle(log_file, max_size_mb=max_size_mb)
    max_size_bytes = int(max_size_mb * 1024 * 1024)
    if os.path.exists(log_file):
        os.remove(log_file)

    try:
        # Run a command that generates a lot of output
        # Use a command that will generate more output than our size limit
        command = "python -c \"for i in range(1000): print('A' * 100)\""

        proc = await asyncio.create_subprocess_shell(
            command, stdout=log_handle, stderr=log_handle
        )
        time.sleep(1)

        assert os.path.exists(log_file), "Log file should exist"
        file_size = os.path.getsize(log_file)
        print(
            f"Final file size: {file_size} bytes, Max allowed: {max_size_bytes} bytes"
        )
        # Flush and check file size
        log_handle.flush()
        await proc.wait()

        # File size should not exceed max_size_bytes by much (allowing for some overhead from rotation message)
        # We allow some buffer for the rotation message that gets added
        assert file_size > 0, "File size should be greater than 0"
        assert (
            file_size <= max_size_bytes + 1000
        ), f"File size {file_size} exceeds limit {max_size_bytes}"

        # Verify the file contains rotation message if it was rotated
        with open(log_file, "r") as f:
            content = f.read()
            if "=== LOG ROTATED - OLDER ENTRIES REMOVED ===" in content:
                print("Log rotation occurred as expected")

    finally:
        log_handle.close()
        # Clean up test file
        if os.path.exists(log_file):
            os.remove(log_file)


if __name__ == "__main__":
    Fire(
        {
            "truncating": test_truncating_file_handler,
            "rotating": lambda: asyncio.run(test_rotating_file_handle()),
        }
    )



import json


class FlexibleJSONEncoder(json.JSONEncoder):
    def encode(self, obj):
        return super().encode(self._convert_keys(obj))

    def _convert_keys(self, obj):
        if isinstance(obj, dict):
            return {self._convert_key(k): self._convert_keys(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [self._convert_keys(i) for i in obj]
        return obj

    def _convert_key(self, key):
        if isinstance(key, (str, int, float, bool, type(None))):
            return key
        return str(key)

    def default(self, obj):
        try:
            return super().default(obj)
        except TypeError:
            return str(obj)



import base64
import time
import datetime
from copy import deepcopy
import json
import os
from mimetypes import guess_type
from typing import Optional, Type

from openai import AzureOpenAI, OpenAI
from openai._base_client import BaseClient
from openai.types.chat import ChatCompletion

from ykutil.types_util import T
from ykutil.log_util import log
from ykutil.pydantic_models import BatchResponse


# Source: https://learn.microsoft.com/en-us/azure/ai-services/openai/how-to/gpt-with-vision?tabs=rest
# Function to encode a local image into data URL
def local_image_to_data_url(image_path):
    # Guess the MIME type of the image based on the file extension
    mime_type, _ = guess_type(image_path)
    if mime_type is None:
        mime_type = "application/octet-stream"  # Default MIME type if none is found

    # Read and encode the image file
    with open(image_path, "rb") as image_file:
        base64_encoded_data = base64.b64encode(image_file.read()).decode("utf-8")

    # Construct the data URL
    return f"data:{mime_type};base64,{base64_encoded_data}"


def human_readable_parse(messages: list[dict[str, str]]):
    return "\n".join([f'{msg["role"]}:\n{msg["content"]}' for msg in messages])


class ModelWrapper:
    def __init__(
        self, client: BaseClient, model_name: str, log_file: Optional[str] = None
    ):
        self.client = client
        self.model_name = model_name
        self.log_file = log_file
        self.stats = {"requests": 0, "input_tokens": 0, "completion_tokens": 0}

    def complete(self, messages: list[dict[str, str]], **kwargs) -> str:

        response: ChatCompletion = self.client.chat.completions.create(
            model=self.model_name, messages=messages, **kwargs
        )
        self.stats["requests"] += 1
        self.stats["input_tokens"] += response.usage.prompt_tokens
        self.stats["completion_tokens"] += response.usage.completion_tokens
        if self.log_file is not None:
            with open(self.log_file, "a") as f:
                f.write(json.dumps(response.model_dump()) + ",\n")
        out = response.choices[0].message.content
        if out is None:
            log("Bad response", response, level="warn")
        return out

    def structured_complete(
        self, messages: list[dict[str, str]], structure_class: Type[T], **kwargs
    ):
        response = self.client.beta.chat.completions.parse(
            model=self.model_name,
            messages=messages,
            response_format=structure_class,
        )
        self.stats["requests"] += 1
        self.stats["input_tokens"] += response.usage.prompt_tokens
        self.stats["completion_tokens"] += response.usage.completion_tokens
        if self.log_file is not None:
            msg_copy = messages.copy()
            msg_copy.append(response.choices[0].message.dict())
            with open(self.log_file, "a") as f:
                f.write(json.dumps(msg_copy) + ",\n")

        out = response.choices[0].message
        if out is None:
            log("Bad response", response, level="warn")
        return out

    def compute_cost():
        raise NotImplementedError()


class OpenAIModelWrapper(ModelWrapper):
    def __init__(self, model_name: str, api_key: str):
        self.client = OpenAI(api_key=api_key)
        super().__init__(self.client, model_name)


class AzureModelWrapper(ModelWrapper):
    def __init__(
        self,
        model_name: str = "gpt-5",
        log_file=None,
        api_version="2024-12-01-preview",
    ):
        self.client = AzureOpenAI(
            api_key=os.environ.get("AZURE_OPENAI_API_KEY"),
            api_version=api_version,
            azure_endpoint=os.getenv("AZURE_OPENAI_ENDPOINT"),
        )
        super().__init__(self.client, model_name, log_file)

    def compute_cost(
        self,
        input_token_cost: Optional[float] = None,
        output_token_cost: Optional[float] = None,
    ) -> float:
        if input_token_cost is None:
            assert output_token_cost is None
            if "gpt-4o" in self.model_name:
                if "mini" in self.model_name:
                    input_token_cost = 0.000165 / 1000
                    output_token_cost = 0.00066 / 1000
                else:
                    if "2024-08-06" in self.model_name:
                        input_token_cost = 0.0025 / 1000
                        output_token_cost = 0.010 / 1000
                    else:
                        input_token_cost = 0.005 / 1000
                        output_token_cost = 0.015 / 1000
            elif "gpt-35" in self.model_name:
                input_token_cost = 0.0005 / 1000
                output_token_cost = 0.0015 / 1000
            elif "gpt-4.1-nano" in self.model_name:
                input_token_cost = 0.1 / 1_000_000
                output_token_cost = 0.4 / 1_000_000
            elif "gpt-4.1-mini" in self.model_name:
                input_token_cost = 0.4 / 1_000_000
                output_token_cost = 1.6 / 1_000_000
            elif "gpt-4.1" in self.model_name:
                input_token_cost = 2 / 1_000_000
                output_token_cost = 8 / 1_000_000
            elif "gpt-5" in self.model_name:
                input_token_cost = 1.25 / 1000_000
                output_token_cost = 10 / 1000_000
            else:
                raise ValueError(
                    f"Unknown model name: {self.model_name}. Please provide"
                    " input_token_cost and output_token_cost."
                )

        total_cost = (
            self.stats["input_tokens"] * input_token_cost
            + self.stats["completion_tokens"] * output_token_cost
        )
        return total_cost


class BatchAPIWrapper:
    def __init__(
        self,
        model_name,
        api_version=None,
        tmp_dir="~/tmp",
        temperature=1.0,
        custom_content_filter: Optional[str] = None,
    ):
        self.client = AzureOpenAI(
            api_key=os.environ.get("AZURE_OPENAI_API_KEY"),
            api_version=api_version,
            azure_endpoint=os.getenv("AZURE_OPENAI_ENDPOINT"),
        )
        self.model_name = model_name
        self.job_json = {
            "custom_id": None,
            "method": "POST",
            "url": "/chat/completions",
            "body": {"model": self.model_name, "messages": None},
            "temperature": temperature,
        }
        if custom_content_filter is not None:
            self.job_json["headers"] = {"x-policy-id": custom_content_filter}

        self.temp_dir = os.path.expanduser(tmp_dir)

    def create_batch_job(self, many_messages: list[list[dict[str, str]]]):
        with open(os.path.join(self.temp_dir, "batch_job.jsonl"), "w") as f:
            for i, messages in enumerate(many_messages):
                job = deepcopy(self.job_json)
                job["body"]["messages"] = messages
                job["custom_id"] = str(i)
                f.write(json.dumps(job) + "\n")

        file = self.client.files.create(
            file=open(os.path.join(self.temp_dir, "batch_job.jsonl"), "rb"),
            purpose="batch",
        )
        batch_response = self.client.batches.create(
            input_file_id=file.id,
            endpoint="/chat/completions",
            completion_window="24h",
        )

        return batch_response.id

    def get_result(self, batch_id) -> list[BatchResponse]:
        status = "validating"
        while status not in ("completed", "failed", "canceled"):
            batch_response = self.client.batches.retrieve(batch_id)
            status = batch_response.status
            print(f"{datetime.datetime.now()} Batch Id: {batch_id},  Status: {status}")
            if status not in ("completed", "failed", "canceled"):
                time.sleep(60)

        if batch_response.status == "failed":
            for error in batch_response.errors.data:
                print(f"Error code {error.code} Message {error.message}")
        output_file_id = batch_response.output_file_id

        if not output_file_id:
            output_file_id = batch_response.error_file_id

        responses = []

        if output_file_id:
            file_response = self.client.files.content(output_file_id)
            raw_responses = file_response.text.strip().split("\n")

            for raw_response in raw_responses:
                json_response = json.loads(raw_response)
                responses.append(json_response)
                formatted_json = json.dumps(json_response, indent=2)
                with open(os.path.join(self.temp_dir, "batch_results.jsonl"), "a") as f:
                    f.write(formatted_json + "\n")

        return [BatchResponse.model_validate(x) for x in responses]



from typing import Generator, TypeVar

T = TypeVar("T")
U = TypeVar("U")

SimpleGenerator = Generator[T, None, None]


def describe_type(t: object) -> str:
    import numpy as np
    from torch import Tensor

    if isinstance(t, Tensor):
        return "tensor." + str(t.dtype)
    elif isinstance(t, np.ndarray):
        return "ndarray." + str(t.dtype)

    return str(type(t))



from pydantic import BaseModel, ConfigDict
from typing import Any, Dict, List, Optional


# ── Low‑level helpers ──────────────────────────────────────────────────────────
class FilterCategory(BaseModel):
    filtered: bool
    severity: Optional[str] = None  # e.g. "safe", "low", "medium", …
    detected: Optional[bool] = None  # only present on jailbreak

    model_config = ConfigDict(extra="allow")  # tolerate future keys


class ContentFilterResults(BaseModel):
    hate: Optional[FilterCategory] = None
    self_harm: Optional[FilterCategory] = None
    sexual: Optional[FilterCategory] = None
    violence: Optional[FilterCategory] = None
    jailbreak: Optional[FilterCategory] = None

    model_config = ConfigDict(extra="allow")


class Message(BaseModel):
    role: str
    content: Optional[str] = None
    function_call: Optional[Dict[str, Any]] = None
    tool_calls: Optional[List[Dict[str, Any]]] = None
    name: Optional[str] = None  # for function messages, etc.

    model_config = ConfigDict(extra="allow")


# ── Mid‑level wrappers ─────────────────────────────────────────────────────────
class Choice(BaseModel):
    index: int
    message: Optional[Message] = None  # absent in streaming deltas
    delta: Optional[Message] = None  # present in streaming deltas
    finish_reason: Optional[str] = None
    logprobs: Optional[Any] = None
    content_filter_results: Optional[ContentFilterResults] = None

    model_config = ConfigDict(extra="allow")


class PromptFilterResult(BaseModel):
    prompt_index: int
    content_filter_results: ContentFilterResults

    model_config = ConfigDict(extra="allow")


class Usage(BaseModel):
    prompt_tokens: Optional[int] = None
    completion_tokens: Optional[int] = None
    total_tokens: Optional[int] = None

    model_config = ConfigDict(extra="allow")


# ── Top‑level “body” for chat completions ──────────────────────────────────────
class Body(BaseModel):
    id: str
    created: int  # Unix‑epoch seconds
    model: str
    object: str
    choices: List[Choice]
    prompt_filter_results: Optional[List[PromptFilterResult]] = None
    system_fingerprint: Optional[str] = None
    usage: Optional[Usage] = None

    model_config = ConfigDict(extra="allow")


# ── Envelope that most SDKs wrap around the body ──────────────────────────────
class OpenAIResponse(BaseModel):
    body: Body
    request_id: Optional[str] = None
    status_code: int

    model_config = ConfigDict(extra="allow")


class BatchResponse(BaseModel):
    custom_id: str
    response: OpenAIResponse
    error: Optional[str] = None



from datasets import Dataset, DatasetDict
from transformers import PreTrainedTokenizer

from ykutil.constants import IGNORE_INDEX
from ykutil.python import count_sublist_occurrences
from ykutil.transformer import load_tk_with_pad_tk
from ykutil.types_util import describe_type


def count_tokens_in_dataset(ds: Dataset):
    count = 0
    assert "input_ids" in ds.column_names, f"ds.column_names: {ds.column_names}"
    for i in range(len(ds)):
        data = ds[i]
        in_ids = data["input_ids"]
        count += len(in_ids)
    return count


def describe_dataset(ds: Dataset, tokenizer=None, show_rows=(0,), zip_labels=False):
    pr = lambda p: print("###############\n" + p)
    pr("Metadata:")
    print(ds.info)
    pr("Columns:")
    print([{col: describe_type(ds[0][col])} for col in ds.column_names])
    pr("Number of rows:")
    print(len(ds))
    pr("Example rows:")
    for i in range(*show_rows):
        example = ds[i]
        if "input_ids" in example:
            if tokenizer is None:
                print("Input IDs:", example["input_ids"][-40:])
                print("Labels:", example["labels"][-40:])
                # raise ValueError(
                #    "tokenizer is required to show rows for token datasets"
                # )
            else:
                if zip_labels:
                    assert "labels" in example
                    tokens = tokenizer.convert_ids_to_tokens(example["input_ids"])
                    for token, label in zip(tokens, example["labels"]):
                        if label != IGNORE_INDEX:
                            print(f"({token}, {float(label)})", end=" ")
                        else:
                            print(token, end=" ")
                else:
                    example["input_ids"] = tokenizer.decode(example["input_ids"])
                    print(example)
        else:
            for key, value in example.items():
                print(f"{key}: {value}")


def colorcode_example(data: dict, tk: PreTrainedTokenizer, beautify=True, fname=None):
    from termcolor import colored

    data["input_ids"] = data["input_ids"].squeeze()
    data["labels"] = data["labels"].squeeze()
    data["labels"] = data["labels"][data["input_ids"] != tk.eos_token_id]
    data["input_ids"] = data["input_ids"][data["input_ids"] != tk.eos_token_id]
    ignores = data["labels"] == IGNORE_INDEX
    if "weights" in data:
        print("Max weight:", data["weights"].max())
        data["weights"] = data["weights"] / data["weights"].max()
        colors = [
            "red" if x < 0.33 else "red" if x < 0.67 else "red" for x in data["weights"]
        ]
        attr = [
            ["dark"] if x < 0.33 else None if x < 0.67 else ["underline"]
            for x in data["weights"]
        ]

    tok_strs = tk.convert_ids_to_tokens(data["input_ids"])
    tok_strs = [
        (
            x
            if ignores[i]
            else colored(
                x,
                colors[i] if "weights" in data else "red",
                attrs=attr[i] if "weights" in data else None,
            )
        )
        for i, x in enumerate(tok_strs)
    ]
    msg = "".join(tok_strs)
    if beautify:
        msg = (
            msg.replace("▁", " ")
            .replace("<0x0A>", "\n")
            .replace("Ġ", " ")
            .replace("Ċ", "\n")
            .replace("<|reserved_special_token_200|>", "\n")
        )
    if fname is None:
        print(msg)
    else:
        with open(fname, "w") as f:
            f.write(msg)


def colorcode_dataset(
    dd: DatasetDict,
    tk: PreTrainedTokenizer,
    num_start=5,
    num_end=6,
    data_key="train",
    fname=None,
    beautify=True,
):

    for i in range(num_start, num_end):
        data = dd[data_key][i]
        colorcode_example(data, tk, beautify=beautify, fname=fname)


def colorcode_entry(
    token_ds_path,
    fname=None,
    tokenizer_path="mistralai/Mistral-7B-v0.1",
    num_start=0,
    num_end=1,
    beautify=True,
):
    token_dd = DatasetDict.load_from_disk(token_ds_path).with_format("torch")
    tokenizer = load_tk_with_pad_tk(tokenizer_path)
    colorcode_dataset(
        token_dd,
        tokenizer,
        fname=fname,
        num_start=num_start,
        num_end=num_end,
        beautify=beautify,
    )


def count_token_sequence(
    token_ds_path: str, token_sequence: list[int], num_start=0, num_end=1
):
    token_ds = Dataset.load_from_disk(token_ds_path)
    count = 0
    for i in range(num_start, num_end):
        data = token_ds[i]
        in_ids = data["input_ids"]
        count += count_sublist_occurrences(in_ids, token_sequence)

    return count
